"""Client API for users in Nexus."""
from uuid import UUID

import qnexus.exceptions as qnx_exc
from qnexus.client import nexus_client
from qnexus.models.references import UserRef


def get_self() -> UserRef:
    """Get the logged in user."""

    res = nexus_client.get("/api/v6/user/me")

    if res.status_code != 200:
        raise qnx_exc.ResourceFetchFailed(
            message=res.json(), status_code=res.status_code
        )

    user_dict = res.json()

    return UserRef(
        display_name=user_dict["display_name"],
        id=user_dict["id"],
    )


def _fetch(user_id: UUID) -> UserRef:
    """Get a specific user."""

    res = nexus_client.get(f"/api/v6/user/{user_id}")

    if res.status_code != 200:
        raise qnx_exc.ResourceFetchFailed(
            message=res.json(), status_code=res.status_code
        )

    user_dict = res.json()

    return UserRef(
        display_name=user_dict["display_name"],
        id=user_dict["id"],
    )



"""Validation classes for Aer noise models."""
from __future__ import annotations

from typing import TYPE_CHECKING, Annotated, Any, List, Literal, Optional, Tuple, Union
from uuid import UUID, uuid4

from pydantic import BaseModel, Field, field_validator

import qnexus.exceptions as qnx_exc
from qnexus.models.backend_info import Register
from qnexus.models.backend_info import register_from_pytket_node as reg_from_qb
from qnexus.models.backend_info import register_to_pytket_node as reg_to_qb

if TYPE_CHECKING:
    from pytket.extensions.qiskit.backends.crosstalk_model import (
        CrosstalkParams as PytketCrosstalkParams,
    )
    from qiskit.circuit import QuantumCircuit  # type: ignore
    from qiskit_aer.noise.noise_model import NoiseModel  # type: ignore


class QiskitBasicInstruction(BaseModel):
    """Validation model for qiskit instructions without params."""

    name: Literal["id", "x", "y", "z", "reset"]
    qubits: List[int]


class QiskitPauliInstruction(BaseModel):
    """Validation model for qiskit pauli string instructions."""

    name: Literal["pauli"] = "pauli"
    params: List[str]
    qubits: List[int]


class QiskitKrausInstruction(BaseModel):
    """Validation model for qiskit kraus operator instructions."""

    name: Literal["kraus"] = "kraus"
    # List of matrices of complex numbers
    params: List[List[List[List[float]]]]
    qubits: List[int]


QiskitInstruction = Annotated[
    Union[
        QiskitBasicInstruction,
        QiskitPauliInstruction,
        QiskitKrausInstruction,
    ],
    Field(discriminator="name"),
]


class AerQuantumError(BaseModel):
    """Validation model for qiskit-aer's QuantumError class."""

    type: Literal["qerror"] = "qerror"
    id: str = Field(default_factory=lambda: uuid4().hex)
    operations: Optional[List[str]] = Field(default_factory=list)
    instructions: List[List[QiskitInstruction]]
    probabilities: List[float] = Field(min_length=1)
    gate_qubits: List[List[int]]

    @field_validator("id")
    def validate_id(cls, value: Any) -> str:  # pylint: disable=no-self-argument
        """Ensure id is a v4 UUID in hex format."""
        return UUID(value, version=4).hex

    # pylint: disable=too-many-locals
    def noise_ops(self) -> List[Tuple["QuantumCircuit", float]]:
        """Return a list of qiskit Circuit and probability tuples for use in QuantumError."""
        # pylint: disable=import-outside-toplevel
        try:
            # fmt: off
            from qiskit.circuit import QuantumCircuit
            from qiskit.circuit.library import IGate  # type: ignore
            from qiskit.circuit.library import Reset  # type: ignore
            from qiskit.circuit.library import PauliGate, XGate, YGate, ZGate
            from qiskit.quantum_info.operators import Kraus  # type: ignore

            # fmt: on
        except ImportError as err:
            raise qnx_exc.OptionalDependencyError(
                "qiskit-aer is not installed. "
                "Please install qiskit-aer before calling noise_ops "
                "'pip install 'qiskit-aer''"
            ) from err

        noise_ops = []
        for instruction, probability in zip(self.instructions, self.probabilities):
            max_qubits = max(max(op.qubits) for op in instruction) + 1
            circuit = QuantumCircuit(max_qubits)
            for operation in instruction:
                if isinstance(operation, QiskitBasicInstruction):
                    if operation.name == "id":
                        circuit.append(IGate(), qargs=operation.qubits)
                    if operation.name == "x":
                        circuit.append(XGate(), qargs=operation.qubits)
                    if operation.name == "y":
                        circuit.append(YGate(), qargs=operation.qubits)
                    if operation.name == "z":
                        circuit.append(ZGate(), qargs=operation.qubits)
                    if operation.name == "reset":
                        circuit.append(Reset(), qargs=operation.qubits)
                if isinstance(operation, QiskitPauliInstruction):
                    circuit.append(
                        PauliGate(
                            label=operation.params[0],
                        ),
                        qargs=operation.qubits,
                    )
                if isinstance(operation, QiskitKrausInstruction):
                    data = [
                        [
                            [complex(array[0], array[1]) for array in matrix]
                            for matrix in param
                        ]
                        for param in operation.params
                    ]

                    circuit.append(
                        Kraus(data=data).to_instruction(), qargs=operation.qubits
                    )

            noise_ops.append((circuit, probability))

        return noise_ops


class AerReadoutError(BaseModel):
    """Validation model for qiskit-aer's ReadoutError class."""

    type: Literal["roerror"] = "roerror"
    operations: Optional[List[str]] = Field(default_factory=lambda: ["measure"])
    probabilities: List[List[float]] = Field(min_length=1)
    gate_qubits: List[List[int]]


Error = Annotated[Union[AerQuantumError, AerReadoutError], Field(discriminator="type")]


class AerNoiseModel(BaseModel):
    """Validation model for qiskit-aer's NoiseModel class."""

    errors: List[Error]

    def to_native(self) -> NoiseModel:
        """Convert to a qiskit native NoiseModel."""
        # pylint: disable=import-outside-toplevel
        try:
            from qiskit_aer.noise.noise_model import (
                NoiseModel,
                QuantumError,
                ReadoutError,
            )
        except ImportError as err:
            raise ImportError(
                "qiskit-aer is not installed. "
                "Please install qiskit-aer before calling `to_native` "
                "'pip install 'qiskit-aer''"
            ) from err

        noise_model = NoiseModel()
        for error in self.errors:
            if isinstance(error, AerQuantumError):
                for applied_qubits in error.gate_qubits:
                    qerror = QuantumError(noise_ops=error.noise_ops())
                    # pylint: disable=protected-access
                    # We can't set the id on a QuantumError usually.
                    qerror._id = error.id
                    # pylint: disable=protected-access
                    # The QuantumError constructor re-scales probabilities
                    # which we want to avoid for serialization.
                    qerror._probs = error.probabilities
                    noise_model.add_quantum_error(
                        error=qerror,
                        instructions=error.operations,
                        qubits=applied_qubits,
                    )
            if isinstance(error, AerReadoutError):
                for applied_qubits in error.gate_qubits:
                    noise_model.add_readout_error(
                        error=ReadoutError(probabilities=error.probabilities),
                        qubits=applied_qubits,
                    )

        return noise_model


class CrosstalkParams(BaseModel):
    """
    Based on pytket-qiskit's CrosstalkParams model.
    Stores various parameters for modelling crosstalk noise.
    """

    zz_crosstalks: dict[Tuple[Register, Register], float]
    single_q_phase_errors: dict[Register, float]
    two_q_induced_phase_errors: dict[Tuple[Register, Register], Tuple[Register, float]]
    non_markovian_noise: list[Tuple[Register, float, float]]
    virtual_z: bool
    N: float
    gate_times: dict[Tuple[str, Tuple[Register, ...]], float]
    phase_damping_error: dict[Register, float]
    amplitude_damping_error: dict[Register, float]

    @classmethod
    def from_pytket_crosstalk_params(
        cls,
        crosstalk_params: PytketCrosstalkParams,
    ) -> "CrosstalkParams":
        """Construct a CrosstalkParams from a pytket CrosstalkParams."""
        return CrosstalkParams(
            zz_crosstalks={
                (
                    reg_from_qb(qbs[0]),
                    reg_from_qb(qbs[1]),
                ): error
                for qbs, error in crosstalk_params.zz_crosstalks.items()
            },
            single_q_phase_errors={
                reg_from_qb(qb): error
                for qb, error in crosstalk_params.single_q_phase_errors.items()
            },
            two_q_induced_phase_errors={
                (reg_from_qb(qbs[0]), reg_from_qb(qbs[1])): (
                    reg_from_qb(qb_plus_phase[0]),
                    qb_plus_phase[1],
                )
                for qbs, qb_plus_phase in crosstalk_params.two_q_induced_phase_errors.items()
            },
            non_markovian_noise=[
                (
                    reg_from_qb(noise_entry[0]),
                    noise_entry[1],
                    noise_entry[2],
                )
                for noise_entry in crosstalk_params.non_markovian_noise
            ],
            virtual_z=crosstalk_params.virtual_z,
            N=crosstalk_params.N,
            gate_times={
                (
                    gate[0].name,
                    tuple((reg_from_qb(qb) for qb in gate[1])),
                ): time_info
                for gate, time_info in crosstalk_params.gate_times.items()
            },
            phase_damping_error={
                reg_from_qb(qb): error
                for qb, error in crosstalk_params.phase_damping_error.items()
            },
            amplitude_damping_error={
                reg_from_qb(qb): error
                for qb, error in crosstalk_params.amplitude_damping_error.items()
            },
        )

    def to_pytket_crosstalk_params(self) -> PytketCrosstalkParams:
        """Construct a pytket CrosstalkParams from a CrosstalkParams."""
        try:
            # pylint: disable=import-outside-toplevel
            from pytket.circuit import OpType
            from pytket.extensions.qiskit.backends.crosstalk_model import (
                CrosstalkParams as PytketCrosstalkParams,
            )
        except ImportError as err:
            raise qnx_exc.OptionalDependencyError from err
        return PytketCrosstalkParams(
            zz_crosstalks={
                (reg_to_qb(regs[0]), reg_to_qb(regs[1])): error
                for regs, error in self.zz_crosstalks.items()
            },
            single_q_phase_errors={
                reg_to_qb(reg): error
                for reg, error in self.single_q_phase_errors.items()
            },
            two_q_induced_phase_errors={
                (reg_to_qb(regs[0]), reg_to_qb(regs[1])): (
                    reg_to_qb(qb_plus_phase[0]),
                    qb_plus_phase[1],
                )
                for regs, qb_plus_phase in self.two_q_induced_phase_errors.items()
            },
            non_markovian_noise=[
                (reg_to_qb(noise_entry[0]), noise_entry[1], noise_entry[2])
                for noise_entry in self.non_markovian_noise
            ],
            virtual_z=self.virtual_z,
            N=self.N,
            gate_times={
                (
                    OpType.from_name(gate[0]),
                    tuple((reg_to_qb(r) for r in gate[1])),
                ): time_info
                for gate, time_info in self.gate_times.items()
            },
            phase_damping_error={
                reg_to_qb(reg): error for reg, error in self.phase_damping_error.items()
            },
            amplitude_damping_error={
                reg_to_qb(reg): error
                for reg, error in self.amplitude_damping_error.items()
            },
        )



"""Client API for authentication in Nexus."""

import getpass
import time
import webbrowser
from http import HTTPStatus

import httpx
from colorama import Fore
from pydantic import EmailStr

# from halo import Halo
from rich.console import Console
from rich.panel import Panel

import qnexus.exceptions as qnx_exc
from qnexus.client import reload_client
from qnexus.client.utils import consolidate_error, remove_token, write_token
from qnexus.config import get_config

console = Console()

_auth_client = httpx.Client(
    base_url=f"{get_config().url}/auth", timeout=None, verify=get_config().httpx_verify
)


def login() -> None:
    """
    Log in to Quantinuum Nexus using the web browser.

    (if web browser can't be launched, displays the link)
    """

    res = _auth_client.post(
        "/device/device_authorization",
        headers={"Content-Type": "application/x-www-form-urlencoded"},
        data={"client_id": "scales", "scope": "myqos"},
    )

    user_code = res.json()["user_code"]
    device_code = res.json()["device_code"]
    verification_uri_complete = res.json()["verification_uri_complete"]
    expires_in = res.json()["expires_in"]
    poll_interval = res.json()["interval"]

    webbrowser.open(verification_uri_complete, new=2)

    token_request_body = {
        "grant_type": "urn:ietf:params:oauth:grant-type:device_code",
        "device_code": device_code,
        "client_id": "scales",
    }

    print("🌐 Browser log in initiated.")

    console.print(
        Panel(
            f"""
        Confirm that the browser shows the following code and click 'allow device':

                                     {user_code}
        """,
            width=90,
        )
    )

    print(
        "Browser didn't open automatically? Use this link: "
        f"{Fore.BLUE + verification_uri_complete}"
    )

    polling_for_seconds = 0
    while polling_for_seconds < expires_in:
        time.sleep(poll_interval)
        polling_for_seconds += poll_interval
        resp = _auth_client.post(
            "/device/token",
            headers={"Content-Type": "application/x-www-form-urlencoded"},
            data=token_request_body,
        )
        if (
            resp.status_code == HTTPStatus.BAD_REQUEST
            and resp.json().get("error") == "AUTHORIZATION_PENDING"
        ):
            continue
        if resp.status_code == HTTPStatus.TOO_MANY_REQUESTS:
            continue
        if resp.status_code == HTTPStatus.OK:
            resp_json = resp.json()
            write_token("refresh_token", resp_json["refresh_token"])
            write_token(
                "access_token",
                resp_json["access_token"],
            )
            # spinner.stop()
            print(
                f"✅ Successfully logged in as {resp_json['email']} using the browser."
            )
            return
        # Fail for all other statuses
        consolidate_error(res=resp, description="Browser Login")
        # spinner.stop()

        reload_client()

        return
    raise qnx_exc.AuthenticationError("Browser login Failed, code has expired.")


def login_with_credentials() -> None:
    """Log in to Nexus using a username and password."""
    user_name = input("Enter your Nexus email: ")
    pwd = getpass.getpass(prompt="Enter your Nexus password: ")

    _request_tokens(user=user_name, pwd=pwd)

    reload_client()
    print(f"✅ Successfully logged in as {user_name}.")


def logout() -> None:
    """Clear tokens from file system."""
    remove_token("refresh_token")
    remove_token("access_token")
    print("Successfully logged out.")


def _request_tokens(user: EmailStr, pwd: str) -> None:
    """Method to send login request to Nexus auth api and save tokens."""
    body = {"email": user, "password": pwd}
    try:
        resp = _auth_client.post(
            "/login",
            json=body,
        )

        mfa_redirect_uri = resp.json().get("redirect_uri", "")
        if mfa_redirect_uri.startswith("/auth/mfa_challenge/"):
            mfa_code = input("Enter your MFA verification code: ")
            body["code"] = mfa_code
            body.pop("password")
            resp = _auth_client.post(
                "/mfa_challenge",
                json=body,
            )

        terms_redirect_uri = resp.json().get("redirect_uri", "")
        if terms_redirect_uri.startswith("/auth/terms_challenge"):
            message = "Terms and conditions not accepted. To continue, "
            message += "please accept our new terms and conditions by signing in "
            message += "to the Nexus website https://nexus.quantinuum.com/auth/login."

            # logger.error(message)
            raise qnx_exc.AuthenticationError(message)

        _response_check(resp, "Login")

        myqos_oat = resp.cookies.get("myqos_oat", None)
        myqos_id = resp.cookies.get("myqos_id", None)

        if not myqos_oat or not myqos_id:
            raise qnx_exc.AuthenticationError(
                "Authorization cookies missing from response."
            )

        write_token("refresh_token", myqos_oat)
        write_token("access_token", myqos_id)

    finally:
        del user
        del pwd
        del body


def _response_check(res: httpx.Response, description: str) -> None:
    """Consolidate as much error-checking of response"""
    # check if token has expired or is generally unauthorized
    resp_json = res.json()
    if res.status_code == HTTPStatus.UNAUTHORIZED:
        raise qnx_exc.AuthenticationError(
            (
                f"Authorization failure attempting: {description}."
                f"\n\nServer Response: {resp_json}"
            )
        )
    if res.status_code != HTTPStatus.OK:
        raise qnx_exc.AuthenticationError(
            f"HTTP error attempting: {description}.\n\nServer Response: {resp_json}"
        )



"""Client API for backend snapshots in Nexus."""
# Work in progress

# https://staging.myqos.com/api-docs#/backend_snapshots


# def get():
#     pass


# def get_only():
#     pass


# def _fetch():
#     pass


# def update():
#     pass
